import copy
import hashlib
import json
import logging
import pickle
import time
import uuid
from dataclasses import asdict, dataclass
from logging import Logger
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
import tqdm
from torch.optim import Adam
from torch.utils.data import DataLoader, Dataset, random_split

from selcol.traintime.dataset import Chunk, Metadata

logger = logging.getLogger(__name__)


@dataclass
class SelfColDataset(Dataset):
    n_dof: int
    X: torch.Tensor
    Y: torch.Tensor
    header: Metadata

    def __len__(self) -> int:
        return len(self.Y)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, torch.Tensor]:
        return self.X[idx], self.Y[idx]

    @classmethod
    def load(cls, base_path: Path, balance: bool = True) -> "SelfColDataset":
        hash_values = set()

        x_list = []
        y_list = []

        n_dof: Optional[int] = None
        header: Optional[Metadata] = None
        for file_path in base_path.iterdir():
            if not file_path.name.startswith("dataset"):
                continue
            if not file_path.name.endswith(".pkl"):
                continue

            with file_path.open(mode="rb") as f:
                chunk: Chunk = pickle.load(f)
            header = chunk.header
            hash_values.add(hashlib.md5(pickle.dumps(header)).hexdigest())

            for sample in chunk.samples:
                x, y = sample
                x_list.append(torch.from_numpy(x).float())
                y_list.append(torch.tensor(y))

            n_dof = len(chunk.header.joint_names)

        all_hashvalue_same = len(hash_values) == 1
        assert all_hashvalue_same

        X = torch.stack(x_list)
        Y = torch.stack(y_list)

        if balance:
            positive_indices = torch.where(Y > 0.5)[0]
            negative_indices = torch.where(Y <= 0.5)[0]
            n_positive = len(positive_indices)
            n_negative = len(negative_indices)
            logger.info("balancing data... posi: {}. nega: {}".format(n_positive, n_negative))

            if n_positive > n_negative:
                positive_indices = positive_indices[torch.randperm(n_positive)][:n_negative]
            else:
                negative_indices = negative_indices[torch.randperm(n_negative)][:n_positive]

            n_positive = len(positive_indices)
            n_negative = len(negative_indices)
            logger.info("modified to => posi: {}. nega: {}".format(n_positive, n_negative))
            indices = torch.concat([positive_indices, negative_indices])
            X = X[indices]
            Y = Y[indices]

        assert n_dof is not None
        assert header is not None
        return cls(n_dof, X, Y.float(), header)


@dataclass
class SelcolNetConfig:
    n_dof: int
    n_width: int = 200
    n_depth: int = 3


class SelcolNet(nn.Module):
    device: torch.device
    linears: nn.Sequential
    unique_id: str
    config: SelcolNetConfig

    def __init__(self, config: SelcolNetConfig, device: Optional[torch.device] = None):
        super().__init__()
        self._setup_from_config(config)

        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.device = device
        self.unique_id = str(uuid.uuid4())
        self.config = config
        logger.info("model name: {}".format(self.__class__.__name__))
        logger.info("model config: {}".format(config))
        logger.info("model is initialized")

    def put_on_device(self, device: Optional[torch.device] = None):
        if device is not None:
            self.device = device
        self.to(self.device)

    def _setup_from_config(self, config: SelcolNetConfig) -> None:
        layers = []
        layers.append(nn.Linear(config.n_dof, config.n_width))
        layers.append(nn.ReLU())  # type: ignore[arg-type]

        for _ in range(config.n_depth - 1):
            layers.append(nn.Linear(config.n_width, config.n_width))
            layers.append(nn.ReLU())  # type: ignore[arg-type]

        layers.append(nn.Linear(config.n_width, 1))
        layers.append(nn.Sigmoid())  # type: ignore[arg-type]
        self.linears = nn.Sequential(*layers)

    def loss(self, sample: Tuple[torch.Tensor, torch.Tensor]) -> torch.Tensor:
        X, Y = sample
        Y_pred = self.linears(X)
        loss = nn.MSELoss()(Y, torch.squeeze(Y_pred, dim=1))
        return loss

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        return self.linears(inp)


@dataclass
class TrainConfig:
    batch_size: int = 200
    valid_data_ratio: float = 0.1
    learning_rate: float = 0.001
    n_epoch: int = 100


def split_with_ratio(dataset: Dataset, valid_ratio: float = 0.1):
    """split dataset into train and validation dataset with specified ratio"""

    n_total = len(dataset)  # type: ignore
    n_validate = int(valid_ratio * n_total)
    ds_train, ds_validate = random_split(dataset, [n_total - n_validate, n_validate])
    return ds_train, ds_validate


def save_models(model: SelcolNet, dataset_metadata: Metadata, valid_loss: float, base_path: Path):
    directory_name = "pretrained-{}-{}-{}".format(
        dataset_metadata.robot_name, dataset_metadata.urdf_md5sum, model.unique_id
    )
    directory_path = base_path / directory_name
    directory_path.mkdir(exist_ok=True)

    model_copied = copy.deepcopy(model).cpu()

    for n_batch in [1, 8, 16, 32, 64, 128]:
        # dump onnx model for different batch sizes
        n_dof = len(dataset_metadata.joint_names)
        model_file_path = directory_path / "model-{}.onnx".format(n_batch)
        torch.onnx.export(
            model_copied,
            torch.randn(n_batch, n_dof),
            str(model_file_path),
            input_names=["x"],
            output_names=["y"],
        )

    metadata_file_path = directory_path / "metadata.json"
    json_dict = asdict(dataset_metadata)
    json_dict["valid_loss"] = valid_loss
    json_dict["time_stamp"] = time.time()
    json_dict["model_config"] = asdict(model.config)
    with metadata_file_path.open(mode="w") as f:
        json.dump(json_dict, f, indent=2)

    # convert to numpy dict and dump it
    np_layers = []
    for torch_layer in model_copied.linears:
        if isinstance(torch_layer, nn.Linear):
            w = torch_layer.weight.detach().numpy()
            b = torch_layer.bias.detach().numpy()
            params = {"w": w, "b": b}
            np_layers.append({"type": "linear", "params": params})
        elif isinstance(torch_layer, nn.ReLU):
            np_layers.append({"type": "relu", "params": None})
        elif isinstance(torch_layer, nn.Sigmoid):
            np_layers.append({"type": "sigmoid", "params": None})
        else:
            assert False
    numpy_layers_path = directory_path / "numpy_layers.pkl"
    with numpy_layers_path.open(mode="wb") as f:
        pickle.dump(np_layers, f)


def train(model: SelcolNet, dataset: SelfColDataset, config: TrainConfig, base_path: Path):

    dataset_train, dataset_validate = split_with_ratio(dataset, config.valid_data_ratio)

    train_loader = DataLoader(dataset=dataset_train, batch_size=config.batch_size, shuffle=True)
    validate_loader = DataLoader(
        dataset=dataset_validate, batch_size=config.batch_size, shuffle=True
    )

    optimizer = Adam(model.parameters(), lr=config.learning_rate)

    def move_to_device(sample):
        if isinstance(sample, torch.Tensor):
            return sample.to(model.device)
        elif isinstance(sample, list):  # NOTE datalodaer return list type not tuple
            return tuple([e.to(model.device) for e in sample])
        else:
            raise RuntimeError

    valid_loss_mean_min = +np.inf

    for epoch in tqdm.tqdm(range(config.n_epoch)):
        model.train()
        train_loss_list: List[float] = []
        for samples in tqdm.tqdm(train_loader, leave=False):
            optimizer.zero_grad()
            samples = move_to_device(samples)
            train_loss = model.loss(samples)
            train_loss_list.append(train_loss.item())
            train_loss.backward()
            optimizer.step()
        train_loss_mean = float(np.mean(train_loss_list))

        valid_loss_list: List[float] = []
        model.eval()
        for samples in validate_loader:
            samples = move_to_device(samples)
            valid_loss = model.loss(samples)
            valid_loss_list.append(valid_loss.item())
        valid_loss_mean = float(np.mean(valid_loss_list))

        logger.info("epoch: {}".format(epoch))
        logger.info("train loss: {}".format(train_loss_mean))
        logger.info("valid loss: {}".format(valid_loss_mean))

        if valid_loss_mean < valid_loss_mean_min:
            valid_loss_mean_min = valid_loss_mean
            save_models(model, dataset.header, valid_loss_mean_min, base_path)

    return valid_loss_mean_min


def create_default_logger(project_path: Path, prefix: str) -> Logger:
    timestr = "_" + time.strftime("%Y%m%d%H%M%S")
    log_dir_path = project_path / "log"
    log_dir_path.mkdir(parents=True, exist_ok=True)
    log_file_path = log_dir_path / (prefix + timestr + ".log")

    logger = logging.getLogger("selcol")
    logger.setLevel(logging.DEBUG)

    fmt = logging.Formatter("[%(levelname)s] %(asctime)s %(name)s: %(message)s")

    sh = logging.StreamHandler()
    sh.setLevel(logging.INFO)
    sh.setFormatter(fmt)

    fh = logging.FileHandler(str(log_file_path))
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(fmt)

    logger.addHandler(sh)
    logger.addHandler(fh)

    log_sym_path = log_dir_path / ("latest_" + prefix + ".log")

    logger.info("create log symlink :{0} => {1}".format(log_file_path, log_sym_path))
    if log_sym_path.is_symlink():
        log_sym_path.unlink()
    log_sym_path.symlink_to(log_file_path)

    return logger



import logging
import os
import pickle
import uuid
from dataclasses import dataclass
from hashlib import md5
from pathlib import Path
from typing import List, Optional, Set, Tuple
from xml.etree import ElementTree

import numpy as np
import tqdm
import trimesh
from skrobot.model.joint import RotationalJoint
from skrobot.model.robot_model import RobotModel
from skrobot.models.urdf import RobotModelFromURDF

logger = logging.getLogger(__name__)


@dataclass
class CollisionManagerWrap:
    _manager: trimesh.collision.CollisionManager
    _check_link_names: List[str]
    _robot_model: RobotModel
    _ignore_pair_set: Set

    @classmethod
    def from_model(
        cls,
        robot_model: RobotModel,
        ignore_link_names: Optional[List[str]] = None,
        use_visual_mesh: bool = True,
    ) -> "CollisionManagerWrap":
        col_manager = trimesh.collision.CollisionManager()

        def has_ignore_name(name: str) -> bool:
            if ignore_link_names is None:
                return False
            for ignore_name in ignore_link_names:
                if ignore_name in name:
                    return True
            return False

        check_link_names = []
        for link in robot_model.link_list:
            if has_ignore_name(link.name):
                continue

            transform = link.worldcoords().T()

            assert use_visual_mesh

            mesh = link.visual_mesh
            if len(mesh) == 0:
                continue

            union_mesh = mesh[0]
            for i in range(1, len(mesh)):
                union_mesh = trimesh.util.concatenate(union_mesh, mesh[i])

            col_manager.add_object(link.name, union_mesh, transform=transform)
            check_link_names.append(link.name)

        _, ignore_set = col_manager.in_collision_internal(return_names=True)
        return cls(col_manager, check_link_names, robot_model, ignore_set)

    def update(self, robot_model: RobotModel) -> None:
        for link_name in self._check_link_names:
            link = robot_model.__dict__[link_name]
            link.visual_mesh
            transform = link.worldcoords().T()
            self._manager.set_transform(link.name, transform=transform)

    def check_collision(self, ignore: bool = True) -> Set[Tuple[str, str]]:
        _, pairs = self._manager.in_collision_internal(return_names=True)
        if ignore:
            pairs = pairs.difference(self._ignore_pair_set)
        return pairs


def get_random_robot_state(
    robot_model: RobotModel, control_joint_names: Optional[List[str]] = None
) -> np.ndarray:
    if control_joint_names is None:
        control_joint_names = robot_model.joint_names

    angles = []
    for jn in control_joint_names:
        joint = robot_model.__dict__[jn]
        if np.isinf(joint.min_angle) or np.isinf(joint.max_angle):
            assert isinstance(joint, RotationalJoint)
            lb = -np.pi
            ub = np.pi
        else:
            lb = joint.min_angle
            ub = joint.max_angle
        angles.append(np.random.rand() * (ub - lb) + lb)
    return np.array(angles)


@dataclass
class Metadata:
    robot_name: str
    urdf_md5sum: str
    joint_names: List[str]
    ignore_link_names: List[str]


@dataclass
class Chunk:
    header: Metadata
    samples: List[Tuple[np.ndarray, bool]]


@dataclass
class TaskArg:
    cache_path: Path
    urdf_path: Path
    ignore_link_names: List[str]
    joint_names: List[str]
    n_sample: int
    disable_tqdm: bool
    bin_filling: bool = False


# @dataclass
# class BinFillingSampler:
#     evaluator: Callable[[np.ndarray], float]
#     robot_model: RobotModel
#     bins: List[List[np.ndarray]]
#     bin_max_list: List[int]
#     max_failure: int = 1000
#
#     @classmethod
#     def create(
#         cls, f, robot_model: RobotModel, n_sample: int, n_bin: int = 20
#     ) -> "BinFillingSampler":
#         bins = [[] for _ in range(n_bin)]
#         bin_max_list = cls.split_number(n_sample, n_bin)
#         return cls(f, robot_model, bins, bin_max_list)
#
#     @staticmethod
#     def split_number(num, div):
#         return [num // div + (1 if x < num % div else 0) for x in range(div)]
#
#     def value_to_bin_index(self, value: float) -> int:
#         assert value <= 1.0 and value >= 0.0
#         n_bin = len(self.bins)
#         width = 1.0 / n_bin
#         return int(value // width)
#
#     def add(self, idx: int, sample: np.ndarray) -> bool:
#         n_max = self.bin_max_list[idx]
#         already_filled = len(self.bins[idx]) == n_max
#         if already_filled:
#             return False
#         self.bins[idx].append(sample)
#
#         is_filled_now = len(self.bins[idx]) == n_max
#         if is_filled_now:
#             logger.info("bin index {} is filled now".format(idx))
#
#         return True
#
#     def is_terminatd(self) -> bool:
#         for n_max, lst in zip(self.bin_max_list, self.bins):
#             if len(lst) < n_max:
#                 return False
#         return True
#
#     def populate(self):
#         n_total = sum(self.bin_max_list)
#         failure_count = 0
#         with tqdm.tqdm(total=n_total, smoothing=0.0) as pbar:
#             while not self.is_terminatd():
#                 vec = get_random_robot_state(self.robot_model)
#                 self.robot_model.angle_vector(vec)
#                 value = self.evaluator(vec)
#                 bin_idx = self.value_to_bin_index(value)
#                 added = self.add(bin_idx, vec)
#                 if added:
#                     failure_count = 0
#                     pbar.update(1)
#                 else:
#                     failure_count += 1
#                 if failure_count > self.max_failure:
#                     logger.warning("reached max failure count. abort and continue.")
#                     break


def create_raw_dataset_chunk(task_arg: TaskArg):
    assert task_arg.cache_path.exists()
    pid = os.getpid()
    np.random.seed(pid)
    logger.info("set seed to {}".format(pid))
    logger.info("task arg {}".format(task_arg))

    robot_model = RobotModelFromURDF(urdf_file=str(task_arg.urdf_path.expanduser()))
    manager = CollisionManagerWrap.from_model(
        robot_model, ignore_link_names=task_arg.ignore_link_names
    )

    urdf_path = task_arg.urdf_path.expanduser()
    with urdf_path.open(mode="r") as f:
        urdf_string = f.read()
    md5sum = md5(urdf_string.encode("utf8")).hexdigest()
    urdf = ElementTree.fromstring(urdf_string)
    header = Metadata(urdf.attrib["name"], md5sum, task_arg.joint_names, task_arg.ignore_link_names)

    if task_arg.bin_filling:
        assert False
        # cache_basepath = default_cache_basepath()
        # from selcol.neural import load_latest_model  # to avoid circular import

        # model, header = load_latest_model(cache_basepath, "pr2")
        # f, grad = model.as_jit_function()
        # sampler = BinFillingSampler.create(f, robot_model, task_arg.n_sample, n_bin=10)
        # sampler.populate()
        # robot_config_list = []
        # for lst in sampler.bins:
        #     robot_config_list.extend(lst)
    else:
        robot_config_list = []
        for _ in range(task_arg.n_sample):
            vec = get_random_robot_state(robot_model, task_arg.joint_names)
            robot_config_list.append(vec)

    samples: List[Tuple[np.ndarray, bool]] = []
    total_count = 0
    positive_count = 0
    for robot_config in tqdm.tqdm(robot_config_list, disable=task_arg.disable_tqdm):

        # update config
        for angle, jn in zip(robot_config, task_arg.joint_names):
            robot_model.__dict__[jn].joint_angle(angle)
        manager.update(robot_model)

        # check
        pairs = manager.check_collision()
        colliding = len(pairs) > 0

        samples.append((robot_config, colliding))

        total_count += 1
        if colliding:
            positive_count += 1
        logger.debug("positive ratio {}".format(float(positive_count) / float(total_count)))

    file_path = task_arg.cache_path / (
        "dataset-{}-{}dof".format(header.robot_name, len(header.joint_names))
        + str(uuid.uuid4())
        + ".pkl"
    )

    logger.info("chunk positive ratio is {}".format(positive_count / len(samples)))
    logger.info("save chunk to {}".format(file_path))
    with file_path.open(mode="wb") as f:
        pickle.dump(Chunk(header, samples), f)



import os
import uuid
from pathlib import Path


def default_pretrained_basepath() -> Path:
    import selcol

    default_path = Path(selcol.__file__).parent / "pretrained"
    assert default_path.exists()
    return default_path


def default_cache_basepath() -> Path:
    default_path = Path("~/.cache").expanduser() / "selcol"
    default_path.mkdir(exist_ok=True)
    return default_path


def default_datachunk_basepath() -> Path:
    base_path = Path("/tmp") / "selcol"
    chunk_base_path = base_path / str(uuid.uuid4())
    chunk_base_path.mkdir(exist_ok=True, parents=True)
    return chunk_base_path


def latest_detachunk_basepath() -> Path:
    base_path = Path("/tmp") / "selcol"
    dirs = list(base_path.iterdir())
    dirs_sorted = sorted(dirs, key=lambda x: os.stat(x).st_mtime)
    return dirs_sorted[-1]



import json
from abc import ABC, abstractmethod
from hashlib import md5
from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Type, TypeVar, Union

import numpy as np


def determine_load_path(
    base_path_list: Sequence[Path],
    robot_name: Optional[str] = None,
    urdf_path: Optional[Path] = None,
    criterion: str = "time_stamp",
) -> Path:

    dir_path_list = []
    for base_path in base_path_list:
        base_path = base_path.expanduser()
        dir_path_list.extend(base_path.iterdir())

    urdf_md5sum: Optional[str] = None
    if urdf_path is not None:
        urdf_path = urdf_path.expanduser()
        assert urdf_path.exists()
        with urdf_path.open(mode="r") as f:
            urdf_string = f.read()
        urdf_md5sum = md5(urdf_string.encode("utf8")).hexdigest()

    best_dir_path: Optional[Path] = None
    best_score = -np.inf
    for dir_path in dir_path_list:
        if not dir_path.name.startswith("pretrained"):
            continue
        json_path = dir_path / "metadata.json"
        with json_path.open(mode="r") as f:
            json_dict = json.load(f)

        # check robot name match
        if robot_name is not None:
            if json_dict["robot_name"] != robot_name:
                continue

        # check urdf md5sum match
        if urdf_md5sum is not None:
            if json_dict["urdf_md5sum"] != urdf_md5sum:
                continue

        if criterion == "time_stamp":
            score = json_dict["time_stamp"]
        elif criterion == "valid_loss":
            score = json_dict["valid_loss"]
        else:
            assert False, "no such criterion {}".format(criterion)

        if best_score < score:
            best_score = score
            best_dir_path = dir_path

    assert best_dir_path is not None, "no pretrained model found matching condition"

    return best_dir_path


InferencerT = TypeVar("InferencerT", bound="SelColInferencerBase")


class SelColInferencerBase(ABC):
    joint_names: List[str]
    eval_joint_indices: Optional[np.ndarray]  # int
    _context: Optional[np.ndarray]

    @classmethod
    def load(
        cls: Type[InferencerT],
        base_path: Union[Path, Sequence[Path]],
        eval_joint_names: Optional[List[str]] = None,
        robot_name: Optional[str] = None,
        urdf_path: Optional[Path] = None,
        criterion: str = "time_stamp",
    ) -> InferencerT:
        if not isinstance(base_path, Sequence):
            base_path = [base_path]
        load_path = determine_load_path(base_path, robot_name, urdf_path, criterion)
        return cls.load_from_path(load_path, eval_joint_names)

    @classmethod
    @abstractmethod
    def load_from_path(
        cls: Type[InferencerT], dir_path: Path, eval_joint_names: Optional[List[str]] = None
    ) -> InferencerT:
        ...

    @abstractmethod
    def infer(self, q: np.ndarray, with_grad: bool = False) -> Tuple[float, np.ndarray]:
        ...

    def set_context(self, q: np.ndarray) -> None:
        assert len(q) == self.dof
        self._context = q

    @property
    def dof(self) -> int:
        return len(self.joint_names)

    @property
    def eval_dof(self) -> int:
        if self.eval_joint_indices is None:
            return self.dof
        else:
            return len(self.eval_joint_indices)



import copy
import json
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import numpy as np
import onnxruntime as ort

from selcol.runtime.interface import SelColInferencerBase


@dataclass
class OrtSelColInferencer(SelColInferencerBase):
    session_table: Dict[int, ort.InferenceSession]
    joint_names: List[str]
    eval_joint_indices: Optional[np.ndarray]  # int
    _context: Optional[np.ndarray]
    metadata_dict: Dict
    eps_numerical_grad: float = 1e-4

    @classmethod
    def load_from_path(
        cls, dir_path: Path, eval_joint_names: Optional[List[str]] = None
    ) -> "OrtSelColInferencer":
        dir_path = dir_path.expanduser()

        metadata_path = dir_path / "metadata.json"
        with metadata_path.open(mode="r") as f:
            json_dict = json.load(f)
        joint_names: List[str] = json_dict["joint_names"]

        session_table = {}
        for filepath in dir_path.iterdir():
            if filepath.name.endswith(".onnx"):
                match = re.search(r"\d+", filepath.name)
                assert match is not None
                n_batch = int(match.group())
                session = ort.InferenceSession(str(filepath))
                session_table[n_batch] = session

        eval_joint_indices: Optional[np.ndarray] = None
        if eval_joint_names is not None:
            eval_joint_indices = np.array([joint_names.index(name) for name in eval_joint_names])

        return cls(session_table, joint_names, eval_joint_indices, None, json_dict)

    def infer(self, q: np.ndarray, with_grad: bool = False) -> Tuple[float, np.ndarray]:
        assert q.ndim == 1
        if self.eval_joint_indices is not None:
            assert (
                self._context is not None
            ), "if you specified eval joints, you neeed specify context"

        if not with_grad:
            return self._infer_without_grad(q)
        else:
            return self._infer_with_grad(q)

    def _find_session_key(self, n_eval: int) -> int:
        batches = np.array(list(self.session_table.keys()))
        n_batch = np.min(batches[batches >= n_eval])
        return n_batch

    def _infer_without_grad(self, q: np.ndarray) -> Tuple[float, np.ndarray]:
        n_batch = self._find_session_key(1)
        session = self.session_table[n_batch]

        if self.eval_joint_indices is None:
            inp = q.reshape(n_batch, -1)
        else:
            assert self._context is not None
            inp = copy.deepcopy(self._context).reshape(1, -1)
            inp[0][self.eval_joint_indices] = q
        input_dict = {"x": inp.astype(np.float32)}
        out = session.run(None, input_dict)[0].item()
        return float(out), np.empty(0)

    def _infer_with_grad(self, q: np.ndarray) -> Tuple[float, np.ndarray]:
        # because we need to compute numerical gradient ...
        n_batch_required = 1 + self.eval_dof
        n_batch = self._find_session_key(n_batch_required)
        assert n_batch >= n_batch_required
        session = self.session_table[n_batch]

        if self.eval_joint_indices is None:
            inp = np.tile(q, (n_batch, 1))
            for i in range(self.dof):
                inp[i + 1, i] += self.eps_numerical_grad
            input_dict = {"x": inp.astype(np.float32)}
            out = session.run(None, input_dict)[0]
            f0 = out[0].item()
            f1s = out[1 : 1 + self.dof, 0]
            grad = (f1s - f0) / self.eps_numerical_grad
            return f0, grad
        else:
            assert self._context is not None
            q_all = copy.deepcopy(self._context)
            q_all[self.eval_joint_indices] = q
            inp = np.tile(q_all, (n_batch, 1))
            for i, idx in enumerate(self.eval_joint_indices):
                inp[i + 1, idx] += self.eps_numerical_grad
            input_dict = {"x": inp.astype(np.float32)}
            out = session.run(None, input_dict)[0]
            f0 = out[0].item()
            f1s = out[1:n_batch_required, 0]
            grad = (f1s - f0) / self.eps_numerical_grad
            return f0, grad



from setuptools import find_packages, setup

install_requires = ["numpy", "onnxruntime"]
dev_requires = ["torch", "scikit-robot", "tqdm"]

setup(
    name="selcol",
    version="0.0.4.1",
    description="yet another percol implementation",
    author="Hirokazu Ishida",
    author_email="h-ishida@jsk.imi.i.u-tokyo.ac.jp",
    license="MIT",
    install_requires=install_requires,
    extras_require={"dev": dev_requires},
    packages=find_packages(exclude=("tests", "docs")),
    package_data={"selcol": ["py.typed"]},
    include_package_data=True,
)



import copy
import json
import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, List, Optional, Tuple

import flax.linen as fnn
import jax
import jax.numpy as jnp
import numpy as np

from selcol.runtime.interface import SelColInferencerBase


@dataclass
class JaxSelColInferencer(SelColInferencerBase):
    f: Callable[[np.ndarray], float]
    grad: Callable[[np.ndarray], np.ndarray]
    joint_names: List[str]
    eval_joint_indices: Optional[np.ndarray]  # int
    _context: Optional[np.ndarray]
    eps_numerical_grad: float = 1e-4

    @dataclass
    class JaxLinear:
        w: np.ndarray
        b: np.ndarray

        @classmethod
        def from_numpy(cls, w_: np.ndarray, b: np.ndarray):
            w = jax.numpy.transpose(w_, (1, 0))
            return cls(jnp.array(w), jnp.array(b))

        def __call__(self, x):
            return jnp.dot(x, self.w) + self.b

    @classmethod
    def load_from_path(
        cls, dir_path: Path, eval_joint_names: Optional[List[str]] = None
    ) -> "JaxSelColInferencer":
        dir_path = dir_path.expanduser()

        # joint stuff
        metadata_path = dir_path / "metadata.json"
        with metadata_path.open(mode="r") as f:
            json_dict = json.load(f)
        joint_names: List[str] = json_dict["joint_names"]

        eval_joint_indices: Optional[np.ndarray] = None
        if eval_joint_names is not None:
            eval_joint_indices = np.array([joint_names.index(name) for name in eval_joint_names])

        # load and parse flax model
        model_path = dir_path / "numpy_layers.pkl"

        with model_path.open(mode="rb") as f:
            numpy_layers = pickle.load(f)

        flax_layers = []
        for numpy_layer in numpy_layers:
            if numpy_layer["type"] == "linear":
                params = numpy_layer["params"]
                w = params["w"]
                b = params["b"]
                flax_layers.append(cls.JaxLinear.from_numpy(w, b))
            elif numpy_layer["type"] == "relu":
                flax_layers.append(fnn.relu)
            elif numpy_layer["type"] == "sigmoid":
                flax_layers.append(fnn.sigmoid)
        model = fnn.Sequential(flax_layers)

        # convert model to jit function
        dummy_input = jnp.empty((len(joint_names),))
        variables = model.init(jax.random.PRNGKey(0), dummy_input)
        f = jax.jit(lambda x: model.apply(variables, x).reshape(()))
        grad = jax.jit(jax.grad(f))
        f(dummy_input)  # jit compile
        grad(dummy_input)  # jit compile
        return cls(f, grad, joint_names, eval_joint_indices, None)

    def infer(self, q: np.ndarray, with_grad: bool = False) -> Tuple[float, np.ndarray]:
        if self.eval_joint_indices is None:
            tweak_indices = np.arange(self.dof)
        else:
            tweak_indices = self.eval_joint_indices

        if self._context is not None:
            q_all = copy.deepcopy(self._context)
            q_all[tweak_indices] = q
        else:
            q_all = q

        f0 = self.f(q_all)
        if not with_grad:
            return f0, np.empty(0)

        # if use jax grad
        use_jit_grad = False
        if use_jit_grad:
            gradval = self.grad(q_all)[tweak_indices]
        else:
            # numerical grad
            gradval = np.zeros(self.eval_dof)
            for i, idx in enumerate(tweak_indices):
                q_all_plus = copy.deepcopy(q_all)
                q_all_plus[idx] += self.eps_numerical_grad
                f1 = self.f(q_all_plus)
                gradval[i] = (f1 - f0) / self.eps_numerical_grad
        return f0, gradval
